/*
 *  sd.cc
 *  
 *
 *  Created by Joakim Flathagen on 08.04.08.
 *  Copyright 2008 FFI. All rights reserved.
 *
 */


/************* SD class ************/
#include "sd.h"
#include "md5.hh"

using namespace std;

/// Constructor
SD::SD(int node_id,OLSR* parent)
{
	_parent = parent;
	_node_id = node_id;
	_num_hash_functions = HASH_FUNCTIONS; // Sets the default value
}

/// Destructor
SD::~SD()
{
	// Should also elete all instances in _foreignservices
	_foreignServices.clear();
}

/// Get a pointer to the local bloom filter
bitset<BLOOMFILTER_SIZE>* SD::getLocalBloomFilter()
{
	return &_localBloomFilter;
}

///
/// \brief Create a 128-bit bloom filter value from an md5 hash
///
/// \param hash Hash value of service name
/// \param bloomFilter Pointer to the bitset containing the bloom filter
///
void SD::createBloomFiltervalue(string* hash, bitset<BLOOMFILTER_SIZE>* bloomFilter)
{
	int bloomSize = BLOOMFILTER_SIZE;
	
	// Calculate the size of each subpart in the md5 hash
	int width = 32 / _num_hash_functions;
	if(width > 8) width = 8;
	int pointer = 0;
	for(int j=0;j<_num_hash_functions;j++)
	{
		string md5Part = hash->substr(pointer,width);
		pointer+=width;
		
		unsigned long int n = ConvertHash2Int(md5Part);
		int BloomValue = n % BLOOMFILTER_SIZE;
		
		// Set bit in bloomfilter
		bloomFilter->set(BloomValue);
	}
}

///
/// \brief Create a 128-bit bloom filter value from an md5 hash
///
/// \param hash Hash value of service name
/// \param bloomFilter Pointer to the bitset containing the bloom filter
///
bool SD::searchBloomFilter(string* hash, bitset<BLOOMFILTER_SIZE>* bloomFilter)
{
	SearchEntry a;
	
	int bloomSize = BLOOMFILTER_SIZE;
	
	// Calculate the size of each subpart in the md5 hash
	int width = 32 / _num_hash_functions;
	if (width > 8) width = 8;
	int pointer = 0;
	bool serviceFound = false;
	for(int j=0;j<_num_hash_functions;j++)
	{
		string md5Part = hash->substr(pointer,width);
		pointer+=width;
		
		unsigned long int n = ConvertHash2Int(md5Part);
		int BloomValue = n % BLOOMFILTER_SIZE;
		
		// Check bit in bloomfilter
		if(bloomFilter->test(BloomValue)) 
		{
			serviceFound = true;
		}
		else return false;
	}
	return serviceFound;
}

///
/// \brief Add a service request to a bloom filter
///
/// \param serviceName Name of the service
///
void SD::createServiceRequest(string serviceName,bitset<BLOOMFILTER_SIZE>* bloomFilter,int requestNumber)
{

	string md5Hash = getmd5Hash(&serviceName);
	createBloomFiltervalue(&md5Hash,bloomFilter);
	SearchEntry s;
	s.Name = serviceName; 
	s.Number = requestNumber;
	_searchServices.push_back(s); // Add to the services this node seaches for
}

///
/// \brief Called when this node receives a service advertisement
///
/// \param bloomfilter of the service
///
bool SD::uponReceivingAdvertisement(bitset<BLOOMFILTER_SIZE>* bloomFilter, int requestNumber)
{

	start:
	vector<SearchEntry>::const_iterator it;
	int i=0;
	for(it=_searchServices.begin(); it != _searchServices.end(); it++)
	{
		
		string temp = (*it).Name;
		int Number = (*it).Number;
		string md5Hash = getmd5Hash(&temp);
		if(searchBloomFilter(&md5Hash,bloomFilter))
		{
			// Service is found on current node id
			cout << "I searched for service " << temp << " and found it" << endl;
			
			// Remove from searchservices.
			_searchServices.erase(_searchServices.begin() + i);
			// We removed an entry and need to run through the array again
			// In case the advertisement matches two or more requests
			
			// Create trace
			std::stringstream sstr;
			if(requestNumber == Number) 
			{
				// This is a respons to my request
				sstr << "SERVICETRUEPOSITIVE ";
			}
			else
			{
				// This is not a resonse to my request
				// Expect a long delay
				sstr << "SERVICEDELAYEDTRUEPOSITIVE ";
			}
			sstr << temp;
			sstr << " ";
			sstr << Number;
			
			string trace = sstr.str();
			
			_parent->traceSDEvent(&trace);


			goto start;
		}
		i++;
	}
	return false;

}

///
/// \brief Resets all the bits in the local bloom filter
///
void SD::resetLocalBloomFilter()
{
	_localBloomFilter.reset();
}


///
/// \brief Print the content of the local bloom filter
///
void SD::printLocalBloomFilter()
{
	printBloomFilter(&_localBloomFilter);
}


///
/// \brief Search for a service in the local bloom filter
///
bool SD::searchLocalService(string query)
{
	string hash = getmd5Hash(&query);
	return searchBloomFilter(&hash,&_localBloomFilter);
}

///
/// \brief Search for a service in the local bloom filter
///
bool SD::searchLocalService(bitset<BLOOMFILTER_SIZE>* bloomFilter)
{
	bitset<BLOOMFILTER_SIZE> comp = *bloomFilter & _localBloomFilter;
	if(comp == *bloomFilter) return true;
	return false;
}

///
/// \brief Outputs the content of a bloom filter
///
void SD::printBloomFilter(bitset<BLOOMFILTER_SIZE>* bloomFilter)
{
	for(int i=0;i<BLOOMFILTER_SIZE;i++)
	{
		if(bloomFilter->test(i))
		{ 
			cout << "1" ;
		} else cout << "0" ;
		
	}
	cout << endl;
}


///
/// \brief Converts a hex hash value to an integer
///
/// \param hash Output from an independent hash function
///
unsigned long int SD::ConvertHash2Int(string hash)
{
	std::stringstream stream;
	stream << hash;
	unsigned long int n;
	stream >> std::hex >> n;
	return n;
}


///
/// \brief Insert a local service in the service register
///
/// \param service Name of the local service
///
void SD::insertLocalService(string service)
{
	// Insert to bloom filter
	insertLocalServiceToBloom(service);
	// Add the name of the service to the array
	insertLocalServiceName(&service);
}

///
/// \brief Insert a local service in the bloom filter
///
/// \param service Name of the local service
///
void SD::insertLocalServiceToBloom(string service)
{
	// Get the md5 hash value of the service
	string hashValue = getmd5Hash(&service);
	// Add to the bloom filter
	createBloomFiltervalue(&hashValue,&_localBloomFilter);

}

///
/// \brief Remove a local service from the service register
///
/// \param service Name of the local service
///
void SD::removeLocalService(string service)
{
	// It is impossible to remove a service from the bloom filter
	// without a possibility that another service is affected.
	// Hence, the bloom filter must be recalculated upon removal
	// of a service
	
	// a) remove service from list of services
	bool serviceFound = removeLocalServiceName(&service);
	if(!serviceFound) return;
	
	// b) clear the bloom filter
	cout << "Reset local bloom filter" << endl;
	_localBloomFilter.reset();
	
	// c) recalculate Bloom Filter
	vector<string>::const_iterator it;
	for(it=_localServices.begin(); it != _localServices.end(); it++)
	{
	
		string itService = (*it);
		cout << "Recreate bloom filter with " << itService << endl;
		insertLocalServiceToBloom(itService);
	}

	
}

///
/// \brief Insert a local service in an local array
///
/// \param service Name of the local service
///
void SD::insertLocalServiceName(string* service)
{
	
	vector<string>::const_iterator it;
	for(it=_localServices.begin(); it != _localServices.end(); it++)
	{
	
		if((*it) == *service)
		{
			// Service exists - return
			cout << "service exists " << endl;
			return;
		}
	}
	_localServices.push_back(*service);
}

///
/// \brief Insert a local service in an local array
///
/// \param service Name of the local service
///
bool SD::removeLocalServiceName(string* service)
{
	
	vector<string>::const_iterator it;
	int i=0;
	for(it=_localServices.begin(); it != _localServices.end(); it++)
	{
		
		if((*it) == *service)
		{
			// Service exists - return
			cout << "removing service " << endl;
			_localServices.erase(_localServices.begin() + i);
			return true;
		}
		i++;
	}
	return false;

}

///
/// \brief Convert a service name to a md5 hash
///
/// \param service Name of the service
///
string SD::getmd5Hash(string* service)
{
	MD5 context;
	context.update((unsigned char*)service->c_str(),service->size());
	context.finalize();
	char* digest = context.hex_digest();
	
	string line = digest;
	return line;
}

///
/// \brief Search for services in the local cache
///
/// \param serviceName Name of the service
///
nsaddr_t SD::searchForeignService(string serviceName)
{
	nsaddr_t retVal = -1;
	// NOT COMPLETE!!!
	// 1 Check for timeouts and remove them
		// Will do this later
	// 2 Search for service and return
	string md5Hash = getmd5Hash(&serviceName);
	//cout << "Hash ok" << endl;
	
	vector<ServiceFilter*>::const_iterator it;
	for(it=_foreignServices.begin(); it != _foreignServices.end(); it++)
	{
		if(searchBloomFilter(&md5Hash,(*it)->getBloomFilterPtr()))
		{
			cout << "Found service" << endl;
			// Service is found on current node id
			return (*it)->getAddress();
		}
	}
	
	return retVal;
}

int SD::getCachesize()
{
	return _foreignServices.size();
} 

///
/// \brief Insert foreign bloom filter
///
/// \param address Address of the foreign node
/// \param filter The bloom filter
///
void SD::insertForeignFilter(int address, double& valid_time, bitset<BLOOMFILTER_SIZE>* filter)
{
	bool NodeExists = false;
	vector<ServiceFilter*>::const_iterator it;
	for(it=_foreignServices.begin(); it != _foreignServices.end(); it++)
	{
		if((*it)->getAddress() == address)
		{
			// Node exists in array
			NodeExists = true;
			cout << "Node found" << endl;
			// Replace with new filter
			(*it)->getBloomFilterPtr()->reset();
			*(*it)->getBloomFilterPtr() = *filter;
			
			// Update valid-time
			(*it)->setValidTime(valid_time);
		}
	}
	
	if(NodeExists == false)
	{
		cout << "Node not found" << endl;
		// Create node and add to array
		ServiceFilter* foreignFilter = new ServiceFilter();
		*foreignFilter->getBloomFilterPtr() = *filter;
		foreignFilter->setAddress(address);
		foreignFilter->setValidTime(valid_time);
		_foreignServices.push_back(foreignFilter);
	}
}

///
/// \brief Removes all nodes in foreign filter with timeout
///
///
void SD::cleanupForeignFilter(double now)
{
	int i = 0;
	vector<ServiceFilter*>::const_iterator it;
	for(it=_foreignServices.begin(); it != _foreignServices.end(); it++)
	{
		if((*it)->getValidTime() < now)
		{
			cout << "Cleans up " << (*it)->getAddress() << " from local filter" << endl;
			// This advertisement is old - delete it
			_foreignServices.erase(_foreignServices.begin() + i);
		
			// Fix this later. Only one node is removed.
			return;
			
		}
		i++;
		
	}

}
